"""
daemon.py
===========

Contains the daemon functions for the registrar

"""

from typing import Dict

import structlog
import redis

from .registrar import register, deregister
from .config import RegistrarConfig, RouteConfig, RouteMode


logger = structlog.getLogger(__name__)


def run_daemon(config: RegistrarConfig):
    """Run the registration daemon

    Arguments:
        config (RegistarConfig): the root configuration.
    """
    client = redis.Redis(
        host=config.redis_host,
        port=config.redis_port,
        charset="utf-8",
        decode_responses=True
    )

    queue_to_route: Dict[str, RouteConfig] = {
        route_cfg.queue: route_cfg
        for route_cfg in config.routes.values()
    }

    queue_names = list(queue_to_route.keys())
    logger.debug(
        f"waiting on items on redis "
        f"queue{'s' if len(queue_names) > 1 else ''}",
        queues=queue_names,
    )

    while True:
        # fetch an item from the queue to be handled
        queue, value = client.brpop(queue_names)

        try:
            route_cfg = queue_to_route[queue]
            if route_cfg.default_mode == RouteMode.REGISTER:
                register(route_cfg, value)
            elif route_cfg.default_mode == RouteMode.DEREGISTER:
                deregister(route_cfg, value)
            if route_cfg.default_mode == RouteMode.DEREGISTER_IDENTIFIER:
                deregister(route_cfg, value, use_id=True)
        except Exception as exc:
            logger.exception(exc)
